; Scheme 9 from Empty Space, Function Library
; By Nils M Holm, 2009,2010
; See the LICENSE file of the S9fES package for terms of use
;
; (define make-matcher symbol list ...)           ==>  procedure
; (define-matcher symbol <clause> ...)            ==>  unspecific
; (let-matcher symbol (<clause> ...) <expr> ...)  ==>  unspecific
;
; (load-from-library "matcher.scm")
;
; These constructs allow to write programs using pattern matching
; style.
;
; The MAKE-MATCHER procedures creates a matcher from the given
; clauses which are represented by LISTs. Each <clause> must have
; the form:
;
; (<pattern> <expression> ...)
;
; where <pattern> is a Scheme object and <expression> is an
; arbitrary expression. MAKE-MATCHER returns a procedure which
; matches its single argument against each given <pattern> and
; evaluates the expressions associated with the first matching
; pattern.
;
; Patterns are matched as follows:
;
; - Atomic objects except for symbols match themselves;
; - Symbols match any value and bind the symbol to that value;
;   the <expression>s will be evaluated with these bindings in
;   effect;
; - The symbol _ matches any value, but does not bind to it;
; - A quoted symbol matches the symbol being quoted;
; - A list is matched by matching its members recursively.
;
; The SYMBOL passed to MAKE-MATCHER is the name of the resulting
; matcher. It is merely used for error reporting. In order for
; a matcher to recurse, it must be bound using DEFINE-MATCHER
; or LET-MATCHER.
;
; DEFINE-MATCHER binds a matcher to a symbol at the toplevel.
; A procedure resembling LENGTH may be written in this way using
; DEFINE-MATCHER:
;
; (define-matcher len
;   (()      0)               ; The empty list has a length of 0.
;   ((_ . x) (+ 1 (len x))))  ; A list whose car part does not
;                             ; matter and whose cdr part is X
;                             ; has a length of 1 plus the length
;                             ; of X.
;
; LET-MATCHER binds a matcher locally. Here is the above matcher
; as a local definition:
;
; (let-matcher len
;   ((()      0)
;    ((_ . x) (+ 1 (len x))))
;   (len '(a b c)))
;
; A matcher accepts only one single argument, but this argument
; may be a list of objects, thereby emulating multiple arguments
; in the same way as ML. However, you would use LIST or quasiquotation
; to transport multiple arguments in Scheme. See the below examples
; for examples.
;
; Example:   (begin
;              (define-matcher len
;                (()      0)
;                ((_ . x) (+ 1 (len x))))
;              (len '(a b c d e f)))               ==>  6
;
;            (let-matcher fac
;              (((0 r) r)
;               ((x r) (fac (list (- x 1) (* x r))))
;               (x     (fac (list x 1))))
;              (fac 10))                           ==>  3628800
;
;            (let-matcher appnd
;              (((x ())      x)
;               ((() x)      x)
;               (((h . t) r) (cons h (appnd `(,t ,r)))))
;              (appnd '((a b c) (d e f))))         ==>  (a b c d e f)

(load-from-library "group-list.scm")

(define (make-matcher name . clauses)

  (define *accessors* '())

  (define *arg* 'arg)

  (define (cad*r-name a)
    (string->symbol
      (apply string-append
             (map symbol->string
                  (append '(c) a '(r))))))

  (define (make-accessor-call path x)
    (if (null? path)
        x
        (let ((a (cad*r-name path)))
          (if (not (memq a *accessors*))
              (set! *accessors* (cons a *accessors*)))
          `(,a ,x))))

  (define (make-accessor path)
    (let ((acc (make-accessor-call path '())))
      (if (null? acc)
          acc
          (car acc))))

  (define (match-object type-pred eqv-pred accessor pattern)
    (let ((tmp (gensym)))
      `((let ((,tmp ,(make-accessor-call accessor *arg*)))
          (and (,type-pred ,tmp)
               (,eqv-pred ,tmp ,pattern))))))

  (define (single-matcher pattern accessor)
    (cond ((and (pair? pattern)
                (eq? 'quote (car pattern)))
             `((eq? ,pattern ,(make-accessor-call accessor *arg*))))
          ((pair? pattern)
            (let loop ((p pattern)
                       (a accessor)
                       (r '()))
              (cond ((pair? p)
                      (loop (cdr p)
                            (cons 'd a)
                            (cons (append `((pair? ,(make-accessor-call
                                                      a
                                                      *arg*)))
                                          (single-matcher (car p)
                                                          (cons 'a a)))
                                  r)))
                    ((null? p)
                      (apply append (reverse (cons (single-matcher p a) r))))
                    (else
                      (apply append (reverse (cons (single-matcher p a)
                                                   r)))))))
          (else
            (cond ((symbol? pattern)
                    `())
                  ((null? pattern)
                    `((null? ,(make-accessor-call accessor *arg*))))
                  ((number? pattern)
                    (match-object 'number? '= accessor pattern))
                  ((eq? #f pattern)
                    `((not ,(make-accessor-call accessor *arg*))))
                  ((eq? #t pattern)
                    `((eq? #t ,(make-accessor-call accessor *arg*))))
                  ((char? pattern)
                    (match-object 'char? 'char=? accessor pattern))
                  ((string? pattern)
                    (match-object 'string? 'string=? accessor pattern))
                  ((vector? pattern)
                    (match-object 'vector? 'equal? accessor pattern))
                  (else
                    (error "unknown object in pattern" pattern))))))

  (define (fetch-variables pattern accessor)
    (cond ((pair? pattern)
            (let loop ((p pattern)
                       (a accessor)
                       (r '()))
              (cond ((and (pair? p)
                          (eq? 'quote (car p)))
                      (apply append (reverse! r)))
                    ((pair? p)
                      (loop (cdr p)
                            (cons 'd a)
                            (cons (fetch-variables (car p) (cons 'a a))
                                  r)))
                    ((null? p)
                      (apply append (reverse! r)))
                    (else
                      (loop '() a (cons (fetch-variables p a)
                                        r))))))
          (else
            (if (and (symbol? pattern)
                     (not (eq? '_ pattern)))
                `(,(cons pattern (make-accessor accessor)))
                `()))))

  (define (make-case pattern body)
    (let* ((matcher (single-matcher pattern '()))
           (env     (fetch-variables pattern '()))
           (vars    (map car env))
           (args    (map (lambda (x)
                           (if (null? (cdr x))
                               *arg*
                               `(,(cdr x) ,*arg*)))
                         env)))
      `((and ,@matcher)
         ((lambda ,vars ,@body) ,@args))))

  (define (gen-cad*r path)
    (let* ((name (cad*r-name path))
           (body (let loop ((p (group-list 4 path)))
                   (if (null? (cdr p))
                       `(,(cad*r-name (car p)) x)
                       (let ((op (cad*r-name (car p))))
                         `(,op ,(loop (cdr p))))))))
      `(,name (lambda (x) ,body))))

  (define (def-accessor accessor)
    (let* ((acc-chars (symbol->string accessor))
           (k         (string-length acc-chars))
           (acc-path  (map (lambda (x)
                             (string->symbol (string x)))
                           (string->list (substring acc-chars 1 (- k 1))))))
      (if (> (length acc-path) 4)
          (list (gen-cad*r acc-path))
          '())))

  (set! *accessors* '())
  (set! *arg* (gensym))
  (let* ((clauses  (map (lambda (x)
                          (make-case (car x) (cdr x)))
                        clauses))
         (acc-defs (apply append (map def-accessor *accessors*))))
    `(lambda (,*arg*)
       (letrec (,@acc-defs)
         (cond ,@clauses
               (else (error "unmatched pattern" (list ',name ,*arg*))))))))

(define-syntax (let-matcher name clauses . body)
  `(letrec ((,name ,(apply make-matcher name clauses))) ,@body))

(define-syntax (define-matcher name . clauses)
  `(define ,name ,(apply make-matcher name clauses)))
